% IPOL MAIN SCRIPT:
%
% Demos use SPA to deal with non-circulant boundary condition.
%
% Requires the L2-r-L0 Deblur Matlab Toolbox v2.1:
% https://www.researchgate.net/publication/325903515_L2rL0deblurMatlabToolbox2p1
% It is included in this ZIP file.
% It is an implementation of the method described in:
% - J. Portilla, A. Tristan-Vega, I.W. Selesnick, "Efficient and robust
%   image restoration using multiple-feature L2-relaxed sparse analysis
%   priors", IEEE Transactions on Image Processing, vol. 24, no. 12,
%   pp. 5046-5059, Dec 2015.
%
% Includes the original MATLAB function "edgetaper" renamed as "edgetaper_matlab" 
% for execution by Octave. The authors claim no rights to this function. 
%
% IPOL SUBMISSION "Spectral Pre-Adaptation for Restoring Real-World Blurred 
% Images Using Standard Deconvolution Methods"
%
% Version 1.2, June 2022
%
% Copyright (c) 2022
% Javier Portilla <javier.portilla@csic.es>
% Chaoqun Dong <chaodong999@gmail.com>
%
% This program is free software: you can redistribute it and/or modify
% it under the terms of the GNU Affero General Public License as
% published by the Free Software Foundation, either version 3 of the
% License, or (at your option) any later version.
%
% This program is distributed in the hope that it will be useful,
% but WITHOUT ANY WARRANTY; without even the implied warranty of
% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
% GNU Affero General Public License for more details.
%
% You should have received a copy of the GNU Affero General Public License
% along with this program. If not, see <http://www.gnu.org/licenses/>.

%%
clear all;
close all;

pkg load image;
pkg load signal;

cwd = fileparts(mfilename('fullpath'));

addpath(genpath([cwd '/../data']))
addpath(genpath([cwd '/../core']))
addpath(genpath([cwd '/../functions']))
addpath(genpath([cwd '/../L2rL0deblurMatlabToolbox2p1']))

%% check using MATLAB or Octave
is_octave = (exist('OCTAVE_VERSION', 'builtin')~=0);

%% set up
args = argv();
image_path = args{1};
kernelid = int32(str2double(args{2}));
sigma2 = str2double(args{3});
restor = args{4};

if strcmp(restor, 'L2rL0') == 0 && strcmp(restor, 'Wiener') == 0
    error("'restor' should be L2rL0 or Wiener");
end
if kernelid <= 0 || kernelid > 4
    error("'kernelid' should be integer between 1 and 4");
end

% read image
x = imread(image_path);

% define kernel
sigma_w = sqrt(sigma2+1e-9);
if kernelid==1,
    % PSF1, non-separable kernel
    i = meshgrid(-7:7, -7:7); j = i';
    kernel = 1./(i.^2 + j.^2 + 1);
    kernel = kernel/sum(kernel(:));
elseif kernelid==2,
    % PSF2, uniform blur
    kernel = ones(9)/81;
    kernel = kernel/sum(kernel(:));
elseif kernelid==3,
    % PSF3, motion blur
    kernel = ones(9,1);
    kernel = kernel/sum(kernel(:));
elseif kernelid==4,
    % PSF4, oblique blur, non-separable kernel
    kernel = [0 0 0 0 0 0 0;
        0 0 0 0 1 1 1;
        0 0 1 2 3 2 1;
        0 1 3 4 3 1 0;
        1 2 3 2 1 0 0;
        1 1 1 0 0 0 0;
        0 0 0 0 0 0 0];
    kernel = kernel/sum(kernel(:));
end


%% simulate observation
[Ny,Nx] = size(x);   % ideal image size

% make sure the kernel is normalized
kernel = kernel/sum(sum(kernel));
[ny,nx] = size(kernel);  % in our implementation the kernel size should be odd for both dimensions
% half kernel size
Ly = ceil((ny-1)/2);
Lx = ceil((nx-1)/2);

x = double(x);
x = cropx8(x);
[Ny,Nx] = size(x); % update image dimensions
% Now both Nx and Ny are even

% Fourier transform
h = zeros(Ny,Nx);
h(Ny/2+1-Ly:Ny/2+1+Ly,Nx/2+1-Lx:Nx/2+1+Lx) = kernel;
H_f= fft2(fftshift(h));

X_f = fft2(x);
randn('seed',0);
Y_f = H_f.*X_f + fft2(sigma_w*randn(size(x)));
y_simu = real(ifft2(Y_f));

y_obsv = y_simu(Ly+1:end-Ly,Lx+1:end-Lx);  % cut off the invalid boundary pixels

%% set extra boundary extension size
[Ley, Lex] = get_extra_size_2(kernel, Ny, Nx);

% final extended size
NDy = Ny + 2*Ley;
NDx = Nx + 2*Lex;


%% psf in final extended size
h = zeros(NDy,NDx);
h(NDy/2+1-Ly:NDy/2+1+Ly,NDx/2+1-Lx:NDx/2+1+Lx) = kernel;
D_f= fft2(fftshift(h));


%% get PSD of the uncorrputed image
% sigma_x and rho (adapted to the [0 255] range) work well for Wiener
% (both boundary processing and Wiener "optimized" at the same time).
im_power = double(ceil(log(max(y_obsv(:)))/log(2)));
im_range = 2^im_power;
PX_f = get_x_psd(im_range, NDy, NDx);


%% preparation for Wiener filter
% blurred noisy image PSD
PY_f = abs(D_f).^2.*PX_f + sigma_w^2;
% Wiener filter
Hwie_f = PX_f.*conj(D_f)./PY_f;


%% preparation for L2-relaxed-L0 restoration
% using L2-relaxed-L0 deblurring Matlab Toolbox v2.1
% please check the toolbox for details

Niter = 10;  % number of iterations for L2-r-L0 restoration

% Load parameters for L2-r-L0 deblurring with Niter=10:
% Here we have de-activated DTCWT for economy.
reduced = 1; % remove the DTCWT, to make it faster
if reduced
    TILs = 1; DTCWT = 0; LDCT = 1;
    rep = [TILs DTCWT LDCT];
    Niter = 10;
    sig2r01 = 2.7966e3;
    sig2r03 = 4.8965e+03;
    sig2r0_v = [sig2r01 1 sig2r03];
    alpha1 = 9.8578;
    alpha3 = 10.5176;
    alpha_v = [alpha1 1 alpha3];
    beta = 0.6368;
else
    load L2rL0_Niter10_parameters.mat
end

LB = 0;
UB = im_range-1;


%% *****************************************************************************************
% ********     Method One: boundary mirror extension and edgetaper tiling (ET)      ********
% ********                     +     Restoration                                    ********
% ******************************************************************************************

% #############       ET      ###############

z_ET = bound_extension(y_obsv,Ly+Ley,Lx+Lex,'mirror'); % mirror extension
PSF = fspecial('gaussian', 40, 6);
if is_octave
  z_ET  = edgetaper_matlab(z_ET, PSF);  % for Octave
else
  z_ET  = edgetaper(z_ET, PSF);   % for MATLAB
end


% #############     RESTORATION    #############
if strcmp(restor,'Wiener')
    
    % ET
    Z_f_ET = fft2(z_ET);
    x_hat_ET = real(ifft2(Z_f_ET.*Hwie_f));
    x_hat_ET = max(min(x_hat_ET, im_range-1), 0);  % back to image range
    
elseif strcmp(restor,'L2rL0')
    
    fprintf(1,'Info:  Deblurring ET result  ...\n');
    x_hat_ET = deblur_L2relaxedL0(z_ET, rep, D_f, sigma_w, Niter, alpha_v, sig2r0_v, beta, LB, UB);
    
end

% cut off the extras extended pixls
% restoration has the same size as ideal image
x_hat_ET = x_hat_ET(Ley+1:end-Ley,Lex+1:end-Lex);


%% ******************************************************************************************
% *********************           Method Three: SPA + Restoration           *******************
% *******************************************************************************************

% #############      SPA     #############

% mask for SPA
mask = zeros(NDy,NDx);
mask(Ley+Ly+1:end-Ly-Ley,Lex+Lx+1:end-Lx-Lex) = 1;
mask = logical(mask);

% masking and extending
y = zeros(NDy,NDx);
y(Ley+Ly+1:end-Ly-Ley,Lex+Lx+1:end-Lx-Lex) = y_obsv.*mask(Ley+Ly+1:end-Ly-Ley,Lex+Lx+1:end-Lx-Lex);

% get initial guess with extrapol.m
y00_obsv = y;
mu = mean2(y00_obsv.*mask)/mean2(mask);
y00 = (y00_obsv - mu).*mask;
% Initiallize surrounded by zeros
z00 = mu*ones(NDy,NDx);
z00(mask) = y00_obsv(mask);
z0 = extrapol(z00, mask);

% SPA
noiter = 50;
z_SPA = SPA(z0,kernel,mask,PX_f,sigma_w,noiter);


% #############     RESTORATION    #############
if strcmp(restor,'Wiener')
    
    Z_f_SPA = fft2(z_SPA);
    x_hat_SPA = real(ifft2(Z_f_SPA.*Hwie_f));
    x_hat_SPA = max(min(x_hat_SPA, im_range-1), 0);  % back to image range
    
elseif strcmp(restor,'L2rL0')
    
    fprintf(1,'Info:  Deblurring SPA result  ...\n');
    x_hat_SPA = deblur_L2relaxedL0(z_SPA, rep, D_f, sigma_w, Niter, alpha_v, sig2r0_v, beta, LB, UB);
    
end

% cut off the extras extended pixels
% restoration has the same size as ideal image
x_hat_SPA = x_hat_SPA(Ley+1:end-Ley,Lex+1:end-Lex);

figure(7),imagescTSc(x_hat_SPA);title('result of deconvolution on SPA')


%% Oracle
% directly deblurring on the simulation

if strcmp(restor,'Wiener')
    
    % PSD of the uncorrputed image in orginal size
    PX_f = get_x_psd(im_range, Ny, Nx);
    
    PY_f = abs(H_f).^2.*PX_f + sigma_w^2;
    Hwie_f = PX_f.*conj(H_f)./PY_f;
    
    Y_SIMU_f = fft2(y_simu);
    x_hat_Ora = real(ifft2(Y_SIMU_f.*Hwie_f));
    x_hat_Ora = max(min(x_hat_Ora, im_range-1), 0);  % back to image range
    
elseif strcmp(restor,'L2rL0')
    
    fprintf(1,'Info:  Deblurring on simulation to get Oracle result  ...\n');
    x_hat_Ora = deblur_L2relaxedL0(y_simu, rep, H_f, sigma_w, Niter, alpha_v, sig2r0_v, beta, LB, UB);
    
end


%% ISNR
% compute ISNR without the invalid boundary pixels affected by CBC
psnr_simu = psnr(y_simu, x);

psnr_ET = psnr(x_hat_ET, x);
isnr_ET = psnr_ET - psnr_simu

psnr_SPA = psnr(x_hat_SPA, x);
isnr_SPA = psnr_SPA - psnr_simu

psnr_Ora = psnr(x_hat_Ora,x);
isnr_Ora = psnr_Ora - psnr_simu

imwrite(uint8(x), 'x.png')
imwrite(uint8(x_hat_SPA), 'x_hat_SPA.png')
imwrite(uint8(x_hat_ET), 'x_hat_ET.png')
imwrite(uint8(x_hat_Ora), 'x_hat_Ora.png')
imwrite(uint8(y_obsv), 'y_simu.png')

q = 30 / quantile(abs(x_hat_Ora(:) - x_hat_SPA(:)), 0.9);
imwrite(uint8(127 + q * (x_hat_Ora - x_hat_SPA)), 'residuals-x_hat_SPA.png')
imwrite(uint8(127 + q * (x_hat_Ora - x_hat_ET)), 'residuals-x_hat_ET.png')
