 function z_hat = SPA(z_init, h, mask, P_x, sigma_w, noiter, reltol)
% This function implements Algorithm 1.
% SPA completes (extends and/or interpolates) an incomplete blurry observation.  
%
%   "SPA" stands for Spectral Pre-Adaptation. 
%   It does a maximum-likelihood interpolation as a likely guess of the unknown 
%   pixels of an incomplete or extended observation of a blurry image ("y"), based on 
%   consistency with the blurring kernel ("h"), the noise standard deviation
%   ("sigma_w", assumed Gaussian), assumed both known, and a simple AR-1 image model.
%   The known and unknown values of "y" correspond to "TRUE"s (or 1s) and "FALSE"s
%   (or 0s) in the mask "S" array, respectively.
%
%   Z_HAT = SPA(Z_INIT, H, MASK, PX_F, SIGMA_W, NOITER) returns a maximum-likelihood
%   interpolation Z_HAT according to a Gaussian spectral model.
%
%   INPUT:
%     z_init          - an initial guess of the interploated image obtained
%                   by any simple interpolation method (In this code
%                   package we provide "extrapol.m" for this purpose.)
%                   (preferrably in the [0 255] range - otherwise please,
%                   normalize it to that range)
%     h           - blurring kernel
%     mask        - a mask indicating the unknown pixels 
%     P_x        - power spectral density model of the uncorrupted image
%     sigma_w     - standard deviation of noise
%     noiter      - (optional) number of iterations of Conjugate Gradient
%     reltol      - (optional) relative tolerance of Conjugate Gradient
%     
%
%   OUTPUT:
%     z_hat       - the interpolated image
%
%   It is useful for:
%   (A) Boundary handling (by extending the boundaries and imposing circular
%   boundaries condition).
%   (B) Deblurring aliased observations, in which case the known
%   subset is a uniformly sampled version of the complete set of pixels.
%   (C) Deblurring when there is kernel mismatch. 
%   (D) Deblurring blurred objects having supports with complex shape support
%   (e.g., out-of-focus background in presence of in-focus foreground).
%
% IPOL SUBMISSION "Spectral Pre-Adaptation for Restoring Real-World Blurred 
% Images Using Standard Deconvolution Methods"
%
% Version 1.2, June 2022
%
% Copyright (c) 2022
% Javier Portilla <javier.portilla@csic.es>
% Chaoqun Dong <chaodong999@gmail.com>
%
% This program is free software: you can redistribute it and/or modify
% it under the terms of the GNU Affero General Public License as
% published by the Free Software Foundation, either version 3 of the
% License, or (at your option) any later version.
%
% This program is distributed in the hope that it will be useful,
% but WITHOUT ANY WARRANTY; without even the implied warranty of
% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
% GNU Affero General Public License for more details.
%
% You should have received a copy of the GNU Affero General Public License
% along with this program. If not, see <http://www.gnu.org/licenses/>.


%%
%fprintf(1,'Info:  Processing by SPA...\n');
    
mask = logical(mask); % in case the mask was given as numbers, in binary, etc.

% blurring kernel h
[ny,nx] = size(h);   % in our implementations the kernel size should be odd for each dimension

Ly = ceil((ny-1)/2);  % half size of the kernel
Lx = ceil((nx-1)/2);


%% step1 in Algorithm 1

% step1.1 and step1.2 in Algorithm 1
% calculate the mean of observed part and subtract it
mu = mean2(z_init.*mask)/mean2(mask);
y_ext = (z_init - mu).*mask;      % masked to have only the observed pixels

% step1.3 in Algorithm 1
% the blurring kernel, in the Fourier domain.
[NDy, NDx] = size(z_init);
h_ext = zeros(NDy,NDx);
h_ext(NDy/2+1-Ly:NDy/2+1+Ly,NDx/2+1-Lx:NDx/2+1+Lx) = h; % Warning: it assumes Nx, Ny are even numbers!
H = fft2(fftshift(h_ext));  %% FT of the kernel

% step1.4 in Algorithm 1
% PSD model for the (incomplete) observation
P_z = abs(H).^2.*P_x + sigma_w^2;


% step1.5 in Algorithm 1
% part of b(y): F_o*y = (F^*)*(S^T)*y in eq.12 
Y_f = fft2(y_ext);

% step1.6 in Algorithm 1
% part of b(y): F_i.D_pz^-1(F_o*.y), here we have F_i^* = (F^*)*(E^T), so Fi = E.F
q = real(ifft2(Y_f./P_z));  

% step1.7 in Algorithm 1
nmask = ~mask;
b_y = -q(nmask); 


%% step2 in Algorithm 1
z = z_init - mu; % We NEED to remove the mean
z_i = z(nmask);

% Conjugate Gragient method
% default values
if ~exist('noiter'),
    noiter = 100;  % number of iterations of CG
end
if ~exist('reltol'),
    reltol = 1e-12; % relative tolerance of CG 
end

X = zeros(size(z));

[xmin,flag,relres,iter] = mycg(@ matrixAMultiplication,b_y,reltol,noiter,[],z_i);
disp(['CG finishes with flag, iterations, relres: ',num2str(flag),', ',num2str(iter),', ',num2str(relres)]);

z_hat = zeros(size(z));
z_hat(mask) = y_ext(mask);
z_hat(nmask) = xmin;
z_hat = z_hat + mu;

    function g = matrixAMultiplication(x)
        % Implements the multiplication of matix A times a vector of
        % no-observed pixels,
        % by first expanding the vector to full size, then weighting it in the
        % Fourier transform, and then, back in the spatial domain, selecting the 
        % non-observed pixels
        X(nmask) = x;
        aux = real(ifft2(fft2(X)./P_z));
        g = aux(nmask);
    end

end

