% DEMO_LENA:
%
% A demo uses SPA for simultaneous block inpainting and deblurring.
%
% Requires the L2-r-L0 Deblur Matlab Toolbox 2.1:
% https://www.researchgate.net/publication/325903515_L2rL0deblurMatlabToolbox2p1
% It is included in this ZIP file. The toolbox is an implementation of the method 
% described in:
% - J. Portilla, A. Tristan-Vega, I.W. Selesnick, "Efficient and robust
%   image restoration using multiple-feature L2-relaxed sparse analysis
%   priors", IEEE Transactions on Image Processing, vol. 24, no. 12,
%   pp. 5046-5059, Dec 2015.
%
% IPOL SUBMISSION "Spectral Pre-Adaptation for Restoring Real-World Blurred 
% Images Using Standard Deconvolution Methods"
%
% Version 1.2, June 2022
%
% Copyright (c) 2022
% Chaoqun Dong <chaodong999@gmail.com>
% Javier Portilla <javier.portilla@csic.es>
%
% This program is free software: you can redistribute it and/or modify
% it under the terms of the GNU Affero General Public License as
% published by the Free Software Foundation, either version 3 of the
% License, or (at your option) any later version.
%
% This program is distributed in the hope that it will be useful,
% but WITHOUT ANY WARRANTY; without even the implied warranty of
% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
% GNU Affero General Public License for more details.
%
% You should have received a copy of the GNU Affero General Public License
% along with this program. If not, see <http://www.gnu.org/licenses/>.


%%
clear all;
close all;

addpath(genpath('../../../data'))
addpath(genpath('../../../core'))
addpath(genpath('../../../functions'))
addpath(genpath('../../../L2rL0deblurMatlabToolbox2p1'))


%% simulate the observation
%{
% ADMM-UBC gives very good performance on simultaneous deblurring and block
% inpainting. Here we redo one experiment from these papers:
%
% [1]  Mariana S. C. Almeida and M�rio A. T. Figueiredo; "Deconvolving images
%      with unknown boundaries using the alternating direction method of
%      multipliers",IEEE Transactions on Image Processing, vol. 22, No. 8,
%      pp. 3074-3086, 2013.
% [2]  Mariana S. C. Almeida and M�rio A. T. Figueiredo; "Frame-based image
%      deblurring with unknown boundary conditions using the Alternating
%      Direction Method of Multipliers", IEEE Int. Conf. on Image Processing
%      Melbourne, Australia, September, 2013.
%
% The blurring kernel and noise level are the same as in Almeida and
% Figueiredo's experiment.
% The missing pixels are approximately in the same (non-reported) location
% as in the their, and the size of the unknown block is exactly the same.
%}
im = double(imread('../../../data/lena256.bmp'));
[Ny,Nx] = size(im);

x0 = im-min(im(:));
x0 = x0/max(x0(:));   % x0 in the range [0,1]

UB = 205;    % setting the upper bound
x0 = UB*x0;

% define blur
BlurDim = 19 ;  % filter size (squared): 19x19 pixels
BSNR = 60;      % BSNR level (in dB)

% a uniform blurring kernel
ht = ones(BlurDim,BlurDim)/(BlurDim*BlurDim);
fsize = round( (BlurDim-1)/2 );         % fsize - size of each side of the filter
h_full = zeros(Ny,Nx);
h_full(Ny/2+1-fsize:Ny/2+1+fsize,Nx/2+1-fsize:Nx/2+1+fsize) = ht;
h_full = ifftshift(h_full);

%{
% In order to compare our result with Almeida and Figueiredo's, we need to 
% make sure the degradation is identical to the one in Almeida and
% Figueiredo's TIP 2013 paper. So we used their code to generate the
% blurring kernel in our experiment.
% The above implementation is an identical approach but created by us.
% One can download Almeida and Figueiredo's code from
% http://www.lx.it.pt/~mscla/ADMM_UBC.htm
% to run their original implementation as below.
%
% %Blurring filter:
n_filter = 1;   % filter type: 1 -  uniform
                %              2 -  out-of-focus
                %              3 -  linear motion at 135�
                %              4 -  Gaussian
ht2 = build_blur( BlurDim , n_filter );         % constructs the blurring filter (19x19)
h_full2 = ht2h( ht2 , size(x0,1) , size(x0,2));  % filter -> h (256x256 pixels)
%}

% In [1], they considered an observation with size 238x238, pixels affected
% by boundary conditions are discarded. In the 256x256 binary mask these
% pixels are inlcuded as zeros. Please see their papers for the details.
load lena_mask.mat

% Blurring degradation (with realistic BC):
y_blurred = ifft2(fft2(h_full).*fft2(x0));     % cyclic convolution

% Additive noise:
y_blurred_valid = y_blurred(1+fsize:end-fsize,1+fsize:end-fsize);
Py = var(y_blurred_valid(:));
sigma_w = sqrt(Py/10^(BSNR/10));
randn('seed',0);
noise = sigma_w*randn(size(y_blurred));

y_simu = (y_blurred+noise).*lena_mask;
y_obsv = y_simu(1+fsize:end-fsize,1+fsize:end-fsize);

figure(1),imagescTSc(y_obsv);title('simulated obsevation')  


%% set boundary extension size
Ley = 5;     
Lex = 5;

NDy = Ny + 2*Ley;
NDx = Nx + 2*Lex;


%% kernel in Fourier Domain
real_psf = padarray(fftshift(h_full),[Ley,Lex],0,'both');
Real_psf_f = real(fft2((ifftshift(real_psf))));


%% mask processing
mask = zeros(NDy,NDx);
mask(Ley+1:end-Ley,Lex+1:end-Lex) = lena_mask;
mask = logical(mask);
figure(2); imagescTSc(mask); title('unknown pixels mask');


%% get PSD of the uncorrputed image
% sigma_x and rho (adapted to the [0 255] range) work well for Wiener 
% (both boundary processing and Wiener "optimized" at the same time). 
im_power = ceil(log(max(y_obsv(:)))/log(2));
sigma_x = 30*2^im_power/256;    
sig2x = sigma_x^2;
rho = 0.65;

[xc, yc] = meshgrid(1:NDx,1:NDy); 
xc = xc - 1 - NDx/2;
yc = yc - 1 - NDy/2;
u = xc/NDx;
v = yc/NDy;
PX_f = sig2x*4*log(rho)^2./((log(rho)^2 + 4*pi^2*u.^2).*(log(rho)^2 + 4*pi^2*v.^2));
PX_f = fftshift(PX_f);


%% masking and extending
y = zeros(NDy,NDx);
y(Ley+fsize+1:end-fsize-Ley,Lex+fsize+1:end-fsize-Lex) = y_obsv;


%% get initial guess with extrapol.m
y_obsv = y;
mu = mean2(y_obsv.*mask)/mean2(mask);
y00 = (y_obsv - mu).*mask;
% Initiallize surrounded by zeros
% Watch out! Here it is assuming "bytes"
z00 = mu*ones(NDy,NDx);
z00(mask) = y_obsv(mask);
z0 = extrapol(z00, mask);


%% SPA
noiter = 12500;
z_hat = SPA(z0,ht,mask,PX_f,sigma_w,noiter);
figure(3),imagescTSc(z_hat);title('result of SPA')   


%% restoration 
% using L2-relaxed-L0 deblurring Matlab Toolbox v2.1
% please check the toolbox for details

Niter = 10;  % number of iterations for L2-r-L0 restoration

% Load parameters for L2-r-L0 deblurring with Niter=10:
% All three representatios are active.
% From Table I, ConDy 10 (number of iterations) in Portilla TIP2015,
% parameters jointly optimizing SSIM of the 24 test experiments.
load L2rL0_Niter10_parameters.mat

LB = 0;

x_hat = deblur_L2relaxedL0(z_hat, rep, Real_psf_f, sigma_w, Niter, alpha_v, sig2r0_v, beta, LB, UB);
figure(4),imagescTSc(x_hat);title('result of deconvolution')  


%% final result (crop the external boundaries)
final_image = x_hat(Ley+1:end-Ley,Lex+1:end-Lex);
figure(5),imagescTSc(final_image);title('final result')

SNR = 10*log10(mean2(x0.^2)/mean2((final_image-x0).^2))

