% DEMO_COMB:
%
% A demo uses SPA for dealing with blurred objects having supports with 
% arbitrary shapes.
% Here is the case (a real image taken by a professional camera) that in an
% image the Foreground is focused and Background is blurred.
% Camera Canon EOS 5D Mark II, 20 Mpix, CR2 to UINT16 TIFF. Focal length
% 50mm, exposure 1/2500s., ISO 100, f/3.5.
% Many thanks to Isabel Portilla for kindly providing the picture.
%
% Requires the L2-r-L0 Deblur Matlab Toolbox 2.1:
% https://www.researchgate.net/publication/325903515_L2rL0deblurMatlabToolbox2p1
% It is included in this ZIP file. The toolbox is an implementation of the method 
% described in:
% - J. Portilla, A. Tristan-Vega, I.W. Selesnick, "Efficient and robust
%   image restoration using multiple-feature L2-relaxed sparse analysis
%   priors", IEEE Transactions on Image Processing, vol. 24, no. 12,
%   pp. 5046-5059, Dec 2015.
%
% IPOL SUBMISSION "Spectral Pre-Adaptation for Restoring Real-World Blurred 
% Images Using Standard Deconvolution Methods"
%
% Version 1.2, June 2022
%
% Copyright (c) 2022
% Chaoqun Dong <chaodong999@gmail.com>
% Javier Portilla <javier.portilla@csic.es>
%
% This program is free software: you can redistribute it and/or modify
% it under the terms of the GNU Affero General Public License as
% published by the Free Software Foundation, either version 3 of the
% License, or (at your option) any later version.
%
% This program is distributed in the hope that it will be useful,
% but WITHOUT ANY WARRANTY; without even the implied warranty of
% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
% GNU Affero General Public License for more details.
%
% You should have received a copy of the GNU Affero General Public License
% along with this program. If not, see <http://www.gnu.org/licenses/>.



%%
clear all;
close all;

addpath(genpath('../../../data'))
addpath(genpath('../../../core'))
addpath(genpath('../../../functions'))
addpath(genpath('../../../L2rL0deblurMatlabToolbox2p1'))


%% check using MATLAB or Octave
is_octave = (exist('OCTAVE_VERSION', 'builtin')~=0);
if is_octave
  fprintf(1,'Info:  Using Octave  ...\n');
 else
  fprintf(1,'Info:  Using MATLAB  ...\n');
end


%% read image
original_image = imread('../../../data/IMG_2373.tiff');
figure(1),imshow(original_image),title('observation')  


%% estimate the object mask
im1 = sum(original_image,3);
im2 = im1<1.1e5;
im3 = imfill(im2,'holes');
if is_octave
  im4 = bwmorph(im3,'dilate',2);
else
  im4 = bwmorph(im3,'dilate',[2 2]);
end
im5 = 1-imfill(im4,'holes');
mask_image = im5;
clear im1 im2 im3 im4 im5


%% get the average of RGB channel
% here we only process a grayscale image which is the average of RGB channel
im00 = double(mean(original_image,3));
[Ny,Nx] = size(im00);


%% set boundary extension size
Ley = 5;     
Lex = 5;

NDy = Ny + 2*Ley;
NDx = Nx + 2*Lex;


%% estimate degradation parameters
D = 29.5;   % blurring kernel diameter, estimated
sigma_w = 4;  % noise level, estimated

[Real_psf_f_o,Real_psf_f,kernel,Ly,Lx,keveny,kevenx] = disk_kernel(D,Ny,Nx,Ley,Lex);


%% mask processing  
% blur mask to avoid artifacts
blur = [1 2 1]'*[1 2 1]/16;
mask_smooth = conv2(mask_image,blur,'valid');
temp = mask_image;
temp(2:end-1,2:end-1) = mask_smooth;
mask_smooth = temp;

mask_image_back = 1-(mask_smooth<1);   % sharp background mask

% boundary handling mask
mask_boun = zeros(NDy,NDx);
mask_boun(Ley+Ly-keveny+1:end-Ly-Ley,Lex+Lx-kevenx+1:end-Lx-Lex) = 1;

% object mask
mask_back = zeros(NDy,NDx);
mask_back(Ley+1:end-Ley,Lex+1:end-Lex) = mask_image_back;

% generic mask
mask = mask_back & mask_boun;   %knows are ones, unknowns are zeros
figure(2); imagescTSc(mask); title('unknown pixels mask');


%% get PSD of the uncorrputed image
sigma_x = 25;   % estimated
sig2x = sigma_x^2; 
rho = 0.65;

[xc, yc] = meshgrid(1:NDx,1:NDy); 
xc = xc - 1 - NDx/2;
yc = yc - 1 - NDy/2;
u = xc/NDx;
v = yc/NDy;
PX_f = sig2x*4*log(rho)^2./((log(rho)^2 + 4*pi^2*u.^2).*(log(rho)^2 + 4*pi^2*v.^2));
PX_f = fftshift(PX_f);


%% masking and extending
y = zeros(NDy,NDx);
y(Ley+Ly+1:end-Ly-Ley,Lex+Lx+1:end-Lx-Lex) = im00(Ly+1:end-Ly,Lx+1:end-Lx).*mask(Ley+Ly+1:end-Ly-Ley,Lex+Lx+1:end-Lx-Lex);


%% get initial guess with extrapol.m
y_obsv = y;
mu = mean2(y_obsv.*mask)/mean2(mask);
y00 = (y_obsv - mu).*mask;
% Initiallize surrounded by zeros
% Watch out! Here it is assuming "bytes"
z00 = mu*ones(NDy,NDx);
z00(mask) = y_obsv(mask);
z0 = extrapol(z00, mask);


%% SPA
z_hat = SPA(z0,kernel,mask,PX_f,sigma_w);
figure(3),imagescTSc(z_hat);title('result of SPA') 


%% restoration
% using L2-relaxed-L0 deblurring Matlab Toolbox v2.1
% please check the toolbox for details

Niter = 10;   % number of iterations for L2-r-L0 restoration

% Load parameters for L2-r-L0 deblurring with Niter=10:
% All three representations are active.
% From Table I, ConDy 10 (number of iterations) in Portilla TIP2015,
% parameters jointly optimizing SSIM of the 24 test experiments.
load L2rL0_Niter10_parameters.mat

LB = 30000;   %lower bound for the background
UB = 54000;   %upper bound for the background

x_hat = deblur_L2relaxedL0(z_hat, rep, Real_psf_f, sigma_w, Niter, alpha_v, sig2r0_v, beta, LB, UB);
figure(4),imagescTSc(x_hat);title('result of deconvolution')   


%% final result
final_image = x_hat(Ley+1:end-Ley,Lex+1:end-Lex).* mask_smooth + im00.*(1-mask_smooth);  % no sharp edge for the object
figure(5),imagescTSc(final_image);title('final result') 

