% DEMO_PIRATE:
%
% A demo uses SPA for deblurring objects in images having supports with 
% arbitrary shape.
% Here is the case (a realistic simulation) that in an image the Foreground
% is focused and Background is blurred (out of focus).
%
% Requires the L2-r-L0 Deblur Matlab Toolbox 2.1:
% https://www.researchgate.net/publication/325903515_L2rL0deblurMatlabToolbox2p1
% It is included in this ZIP file. The toolbox is an implementation of the method 
% described in:
% - J. Portilla, A. Tristan-Vega, I.W. Selesnick, "Efficient and robust
%   image restoration using multiple-feature L2-relaxed sparse analysis
%   priors", IEEE Transactions on Image Processing, vol. 24, no. 12,
%   pp. 5046-5059, Dec 2015.
%
% IPOL SUBMISSION "Spectral Pre-Adaptation for Restoring Real-World Blurred 
% Images Using Standard Deconvolution Methods"
%
% Version 1.2, June 2022
%
% Copyright (c) 2022
% Chaoqun Dong <chaodong999@gmail.com>
% Javier Portilla <javier.portilla@csic.es>
%
% This program is free software: you can redistribute it and/or modify
% it under the terms of the GNU Affero General Public License as
% published by the Free Software Foundation, either version 3 of the
% License, or (at your option) any later version.
%
% This program is distributed in the hope that it will be useful,
% but WITHOUT ANY WARRANTY; without even the implied warranty of
% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
% GNU Affero General Public License for more details.
%
% You should have received a copy of the GNU Affero General Public License
% along with this program. If not, see <http://www.gnu.org/licenses/>.


%%
clear all;
close all;

addpath(genpath('../../../data'))
addpath(genpath('../../../core'))
addpath(genpath('../../../functions'))
addpath(genpath('../../../L2rL0deblurMatlabToolbox2p1'))


%% read images
background_image = double(imread('../../../data/pirate512.png'));
foreground_image = double(imread('../../../data/hand.png'));
[Ny,Nx] = size(background_image);  % ideal image size

% load mask for foreground in focus object
% hand_mask.png is processed from hand.png by thresholding and 
% morphological operations.
% focused foreground pixels are zeros
mask_image = imread('../../../data/hand_mask.png');   
mask_image = im2double(mask_image);  

% blur mask to avoid artifacts
blur = [1 2 1]'*[1 2 1]/16; 
mask_smooth = conv2(mask_image,blur,'valid');
temp = mask_image;
temp(2:end-1,2:end-1) = mask_smooth;
mask_smooth = temp;


%% set boundary extension size
Ley = 8;        % boundary extension parameter
Lex = 8;
NDy = Ny + 2*Ley;
NDx = Nx + 2*Lex;


%% define blur
sigma_w = 0.3;   %low noise level
D = 9;         % diameter of the blurring kernel, in our implementation it should be an odd number
[Real_psf_f_o,Real_psf_f,kernel,Ly,Lx,keveny,kevenx] = disk_kernel(D,Ny,Nx,Ley,Lex); 


%% simulate the observation
B_f = fft2(background_image);
randn('seed',0);
BY_f = Real_psf_f_o.*B_f + fft2(sigma_w*randn(Ny,Nx)); % observation model for noisy convolved image
b_simu = real(ifft2(BY_f));
y_simu = b_simu.* mask_smooth + foreground_image.*(1-mask_smooth);

y_obsv = y_simu(Ly+1:end-Ly,Lx+1:end-Lx); % cut off invalid boundary pixels

figure(1),imagescTSc(y_obsv);title('simulated observation');   


%% mask processing
mask_image_back = 1-(mask_smooth<1);   % sharp background mask
% boundary handling mask
mask_boun = zeros(NDy,NDx);
mask_boun(Ley+Ly-keveny+1:end-Ly-Ley,Lex+Lx-kevenx+1:end-Lx-Lex) = 1;

% object mask
mask_back = zeros(NDy,NDx);
mask_back(Ley+1:end-Ley,Lex+1:end-Lex) = mask_image_back;

% generic mask 
mask = mask_back & mask_boun;   %knows are ones, unknowns are zeros
figure(2); imagescTSc(mask); title('unknown pixels mask');


%% get PSD of the uncorrputed image
% sigma_x and rho (adapted to the [0 255] range) work well for Wiener 
% (both boundary processing and Wiener "optimized" at the same time). 
im_power = ceil(log(max(y_obsv(:)))/log(2));
sigma_x = 30*2^im_power/256;    
sig2x = sigma_x^2;
rho = 0.65;

[xc, yc] = meshgrid(1:NDx,1:NDy); 
xc = xc - 1 - NDx/2;
yc = yc - 1 - NDy/2;
u = xc/NDx;
v = yc/NDy;
PX_f = sig2x*4*log(rho)^2./((log(rho)^2 + 4*pi^2*u.^2).*(log(rho)^2 + 4*pi^2*v.^2));
PX_f = fftshift(PX_f);


%% masking and extending
y = zeros(NDy,NDx);
y(Ley+Ly+1:end-Ly-Ley,Lex+Lx+1:end-Lx-Lex) = y_obsv.*mask(Ley+Ly+1:end-Ly-Ley,Lex+Lx+1:end-Lx-Lex);


%% get initial guess with extrapol.m
y_obsv = y;
mu = mean2(y_obsv.*mask)/mean2(mask);
y00 = (y_obsv - mu).*mask;
% Initiallize surrounded by zeros
% Watch out! Here it is assuming "bytes"
z00 = mu*ones(NDy,NDx);
z00(mask) = y_obsv(mask);
z0 = extrapol(z00, mask);


%% SPA
z_hat = SPA(z0,kernel,mask,PX_f,sigma_w);
figure(3),imagescTSc(z_hat);title('result of SPA')   


%% restoration
% using L2-relaxed-L0 deblurring Matlab Toolbox v2.1
% please check the toolbox for details

Niter = 10;  % number of iterations for L2-r-L0 restoration

% Load parameters for L2-r-L0 deblurring with Niter=10:
% All three representations are active.
% From Table I, ConDy 10 (number of iterations) in Portilla TIP2015,
% parameters jointly optimizing SSIM of the 24 test experiments.
load L2rL0_Niter10_parameters.mat

x_hat = deblur_L2relaxedL0(z_hat, rep, Real_psf_f, sigma_w, Niter, alpha_v, sig2r0_v, beta);
figure(4),imagescTSc(x_hat);title('result of deconvolution')  


%% final result
final_image= x_hat(Ley+1:end-Ley,Lex+1:end-Lex).* mask_smooth + y_simu.*(1-mask_smooth);  % avoid sharp edge between FG and BG
figure(5),imagescTSc(final_image);title('final result')   


%% SNR only for background without invalid boundary pixels
x_back = double(background_image).*mask_image_back;
final_image_back = final_image.*mask_image_back;
SNR_BG = 10*log10(mean2(x_back(Ly+1:end-Ly,Lx+1:end-Lx).^2)/mean2((final_image_back(Ly+1:end-Ly,Lx+1:end-Lx) - x_back(Ly+1:end-Ly,Lx+1:end-Lx)).^2)) 

