% DEMO_KERNEL_SPAVAR_MISMATCH:
%
% Demos use SPA for forcing a blurred observation to be likely in terms of being
% the convolved result of a modeled kernel then do restoration.
% The real PSF is a spatial variant version of the modoled PSF.
%
% This is a second demo for demonstrating how the Spectral Pre-Adaptation method
% helps to make digital restoration less sensitive, more robust, to model errors.
% In this case, we asume a spatially invariant blur model, but simulate an image with
% a spatially variant blur, to measure how this model error impacts on the quality
% of the image restoration. Having a spatially variant blur is a very common problem,
% in imaging devices (e.g. because the optical quality in the vicinity of the optical axis
% is almost always better than on the periphery).
% We apply it here to the practically interesting case where there is aliasing in the image.
% The fact that there is a strong subsampling (3 out of 4 samples are not
% observed!) provides us with more degrees of freedom for constructing an
% intermediate image "spectrally adapted" to fit the model kernel. 
%
% 1) Simullating a spatially variant blurred image, in this case with a blur
% that increases radially from the center of the image.
% 2) Subsampling it (taking one every two consecutive samples along each dimension)
% 3) Interpolating the missing samples using the SPA method  *assuming that that they
% have been blurred with a spatially invariant kernel*
% (the "model kernel"), which corresponds to the kernel for an intermediate radial value
% 4) Restore the resulting image asuming the model kernel.
% 5) Compare the performance drop (with respect to having a spatially
% invariant simulation with the model kernel) caused by this mismatch when using SPA
% interpolation with the performance drop (same) we would experience when applying
% model restoration (4) directly to the spatially variant blurred simulation (1).
%
% Requires the L2-r-L0 Deblur Matlab Toolbox 2.1:
% https://www.researchgate.net/publication/325903515_L2rL0deblurMatlabToolbox2p1
% It is included in this ZIP file. The toolbox is an implementation of the method 
% described in:
% - J. Portilla, A. Tristan-Vega, I.W. Selesnick, "Efficient and robust
%   image restoration using multiple-feature L2-relaxed sparse analysis
%   priors", IEEE Transactions on Image Processing, vol. 24, no. 12,
%   pp. 5046-5059, Dec 2015.
%
% IPOL SUBMISSION "Spectral Pre-Adaptation for Restoring Real-World Blurred 
% Images Using Standard Deconvolution Methods"
%
% Version 1.2, June 2022
%
% Copyright (c) 2022
% Javier Portilla <javier.portilla@csic.es>
% Chaoqun Dong <chaodong999@gmail.com>
%
% This program is free software: you can redistribute it and/or modify
% it under the terms of the GNU Affero General Public License as
% published by the Free Software Foundation, either version 3 of the
% License, or (at your option) any later version.
%
% This program is distributed in the hope that it will be useful,
% but WITHOUT ANY WARRANTY; without even the implied warranty of
% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
% GNU Affero General Public License for more details. 
%
% You should have received a copy of the GNU Affero General Public License
% along with this program. If not, see <http://www.gnu.org/licenses/>.


%%
clear all;
close all;

addpath(genpath('../../data'))
addpath(genpath('../../core'))
addpath(genpath('../../functions'))
addpath(genpath('../../L2rL0deblurMatlabToolbox2p1'))


%% define blur and noise level
D = 9;          
sigma_w = 0.3; 


%% ideal image
x = double(imread('../../data/lena256.bmp'));
%x = double(imread('barbara512.png'));
[Ny,Nx] = size(x);  % ideal image size


%% preparation for L2-r-L0 restoration
% using L2-relaxed-L0 deblurring Matlab Toolbox v2.1
% please check the toolbox for details

Niter = 10;  % number of iterations for L2-r-L0 restoration

% Load parameters for L2-r-L0 deblurring with Niter=10:
% All three representations are active.
% From Table I, ConDy 10 (number of iterations) in Portilla TIP2015,

% parameters jointly optimizing SSIM of the 24 test experiments.
load L2rL0_Niter10_parameters.mat

LB = 0;
UB = 255;


%% Approximate the local blurring kernel as a linear combination of base kernels
% (deformable filtering).
% We construct a vector LUT providing the weights for different diameter D
% values. Then we interpolate this coefficients for intermediate D(r),
% where r is the radial location (the corresponding ring) on the image
% and apply them to the output of filtering with the basis filters.

Nm = 21; % Nm x Nm kernel's spatial support
Np2 = (Nm+1)/2;
cMin = 0.5; 
cMax = 2.1; 
cStep = 0.05; 
c = cMin:cStep:cMax;
Ndiam = length(c);
kernels = zeros(Nm,Nm,Ndiam);
ndiam = 0;
for coef = cMin:cStep:cMax
    d = coef*D;
    [~,~,modeled_kernel,Ly,Lx] = disk_kernel(d,Ny,Nx,0,0);  
    ndiam = ndiam +1;
    kernels(Np2-Ly:Np2+Ly,Np2-Lx:Np2+Lx,ndiam) = modeled_kernel;
end % for coef

kernels = reshape(kernels,Nm^2,Ndiam);
[U,S,V] = svd(kernels.');
Nb = 27;

% Define the basis functions and the interpolating functions
BasisF = V(:,1:Nb);
Alpha = U*S(:,1:Nb);

% Prepare the image filtered with the basis functions
Lm = (Nm-1)/2;
imFiltered = zeros(Ny,Nx,Nb);
for nb = 1:Nb
    bKernel = reshape(BasisF(:,nb),Nm,Nm);
    im_ex = bound_extension(x,Lm,Lm,'circular');
    imFiltered(:,:,nb) = conv2(im_ex,bKernel,'valid');
end    

% Define the radial position on the Nm x Nm grid
R = Ny/2-1;
[X,Y] = meshgrid(-Ny/2:Ny/2-1,-Nx/2:Nx/2-1);
r = sqrt(X.^2 + Y.^2);

%% loop for different percentages of mismatch

[modeled_psf_f_o,modeled_psf_f,modeled_kernel,Ly,Lx] = disk_kernel(D,Ny,Nx,0,0);  % modeled shift invariant kernel
 
noise = sigma_w*randn(Ny,Nx);

mis_percent_list = [];
SNR_SPA_list = [];
SNR_IR_list = [];
SNR_simu_list = [];

for mis_percent = 0:40

    mis_percent_list = [mis_percent_list;mis_percent];

    %% simulate the observation
    coef = 1- (1 - 2*r/R)*mis_percent/100;
    d = coef*D; % local kernel's diameter
    
    % interpolation weights
    W = interp1(c, Alpha, coef,'spline');
    % For a porcentual variation from -mis_percent (in R==0) to +mispercent
    % (in R = Rmax inside the image - corners excluded), we compute the
    % diameter of the kernel for every pixel location
    
    % Spatially variant blurred simulation
    y_simu = sum(imFiltered.*W,3) + noise;
    %figure(1), imagescTSc(y_simu); title('simulation in HR')
    
    y_obsv = y_simu(1:2:end,1:2:end);
    %figure(2), imagescTSc(y_obsv); title('simulated observation in LR')
    
    %% get PSD of the uncorrputed image
    % sigma_x and rho (adapted to the [0 255] range) work well for Wiener 
    % (both boundary processing and Wiener "optimized" at the same time). 
    im_power = ceil(log(max(y_obsv(:)))/log(2));
    sigma_x = 30*2^im_power/256;    
    sig2x = sigma_x^2;
    rho = 0.65;

    [xc, yc] = meshgrid(1:Nx,1:Ny); 
    xc = xc - 1 - Nx/2;
    yc = yc - 1 - Ny/2;
    u = xc/Nx;
    v = yc/Ny;
    PX_f = sig2x*4*log(rho)^2./((log(rho)^2 + 4*pi^2*u.^2).*(log(rho)^2 + 4*pi^2*v.^2));
    PX_f = fftshift(PX_f);
    
    
    %%  *********************************************************
    % ********     SPA + restoration with modeled PSF     *******
    % ***********************************************************
    fprintf(1,'# SPA + Restoration (with modeled PSF)   ...\n');

    % mask processing
    mask = zeros(Ny,Nx);
    mask(1:2:end,1:2:end) = 1;
    mask = logical(mask);

    % masking and extending
    y = zeros(Ny,Nx);
    y(1:2:end,1:2:end) = y_obsv.*mask(1:2:end,1:2:end);

    % get initial guess with extrapol.m
    y00_obsv = y;
    mu = mean2(y00_obsv.*mask)/mean2(mask);
    y00 = (y00_obsv - mu).*mask;
    % Initiallize surrounded by zeros
    z00 = mu*ones(Ny,Nx);
    z00(mask) = y00_obsv(mask);
    z0 = extrapol(z00, mask);

    % SPA
    z_SPA = SPA(z0,modeled_kernel,mask,PX_f,sigma_w);
    %figure(3), imagescTSc(z_SPA); title('result of SPA')

    % L2-rL0 resotration
    x_hat_SPA = deblur_L2relaxedL0(z_SPA, rep, modeled_psf_f_o, sigma_w, Niter, alpha_v, sig2r0_v, beta, LB, UB);
    %figure(4), imagescTSc(x_hat_SPA); title('result of deconvolution on SPA')  

    
    %% *****************************************************************************************
    % **************       Interpolation + Restoration (I+R) with modeled PSF       ************
    % **********************************************************************%*******************
    fprintf(1,'# Spline Interpolation + Restoration (with modeled PSF)   ...\n');

    % preparation for spline interpolation
    [Nsy,Nsx] = size(y_obsv);
    [xsc, ysc] = meshgrid(1:Nsx,1:Nsy);
    [xc, yc] = meshgrid(1:0.5:Nsx+1,1:0.5:Nsy+1);
    xc = xc(1:end-1,1:end-1);
    yc = yc(1:end-1,1:end-1);
    
    % #############    INTERPOLATION   #############
    z_IR = interp2(xsc,ysc,double(y_obsv),xc,yc,'spline');
    %figure(5), imagescTSc(z_IR); title('result of spline interploation')

    % #############     RESTORATION    ##############
    x_hat_IR = deblur_L2relaxedL0(z_IR, rep, modeled_psf_f_o, sigma_w, Niter, alpha_v, sig2r0_v, beta, LB, UB);
    %figure(6), imagescTSc(x_hat_IR); title('result of I + R')
    
    
      %%  ***************************************************************************************************
    % ********     directly restore full simulation with modeled (shift invariant) PSF           ********
    % *****************************************************************************************************

    x_hat_simu = deblur_L2relaxedL0(y_simu, rep, modeled_psf_f_o, sigma_w, Niter, alpha_v, sig2r0_v, beta, LB, UB);
    %figure(7), imagescTSc(x_hat_simu); title('result of direct deconvolution on simulation') 


    %% calculate SNR
    SNR_SPA = 10*log10(mean2(x.^2)/mean2((x_hat_SPA - x).^2)) 
    SNR_IR = 10*log10(mean2(x.^2)/mean2((x_hat_IR - x).^2))
    SNR_simu = 10*log10(mean2(x.^2)/mean2((x_hat_simu - x).^2)) 
    
    SNR_SPA_list = [SNR_SPA_list; SNR_SPA];
    SNR_IR_list = [SNR_IR_list; SNR_IR];
    SNR_simu_list = [SNR_simu_list; SNR_simu];
    


end


%% plot
SNR_SPA_oracle = max(SNR_SPA_list);
SNR_IR_oracle = max(SNR_IR_list);
SNR_simu_oracle = max(SNR_simu_list);

figure(8)
plot(mis_percent_list(1:41),SNR_simu_list(1:41),'ro-',mis_percent_list(1:41),SNR_SPA_list(1:41),'bo-',mis_percent_list(1:41),SNR_IR_list(1:41),'ko-')
grid on;
legend('not aliased (wo/ SPA)', 'aliased (w/ SPA)', 'aliased (wo/ SPA)');
title('Performance effect of the spatially variant blur');
ylabel('SNR drop (dB)')
xlabel('Percentage of kernel''s size variability');
set(gca,'fontsize',20);

h = gca;
set(gcf,'PaperUnits','inches','PaperPosition',[0 0 7 8])

h = xlabel('Percentage of maximum kernel diameter variability');
set(h,'FontSize',20);
h = ylabel('SNR(dB)');
set(h,'FontSize',20);
