% DEMO_KERNEL_SIZE_MISMATCH:
%
% Demos use SPA for forcing a blurred observation to be likely in terms of being
% the convolved result of a modeled kernel then do restoration. 
% The modeled PSF has a slightly different size from the real PSF.
%
% Requires the L2-r-L0 Deblur Matlab Toolbox 2.1:
% https://www.researchgate.net/publication/325903515_L2rL0deblurMatlabToolbox2p1
% It is included in this ZIP file. The toolbox is an implementation of the method 
% described in:
% - J. Portilla, A. Tristan-Vega, I.W. Selesnick, "Efficient and robust
%   image restoration using multiple-feature L2-relaxed sparse analysis
%   priors", IEEE Transactions on Image Processing, vol. 24, no. 12,
%   pp. 5046-5059, Dec 2015.
%
% IPOL SUBMISSION "Spectral Pre-Adaptation for Restoring Real-World Blurred 
% Images Using Standard Deconvolution Methods"
%
% Version 1.2, June 2022
%
% Copyright (c) 2022
% Chaoqun Dong <chaodong999@gmail.com>
% Javier Portilla <javier.portilla@csic.es>
%
% This program is free software: you can redistribute it and/or modify
% it under the terms of the GNU Affero General Public License as
% published by the Free Software Foundation, either version 3 of the
% License, or (at your option) any later version.
%
% This program is distributed in the hope that it will be useful,
% but WITHOUT ANY WARRANTY; without even the implied warranty of
% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
% GNU Affero General Public License for more details.
%
% You should have received a copy of the GNU Affero General Public License
% along with this program. If not, see <http://www.gnu.org/licenses/>.


%%
clear all;
close all;

addpath(genpath('../../data'))
addpath(genpath('../../core'))
addpath(genpath('../../functions'))
addpath(genpath('../../L2rL0deblurMatlabToolbox2p1'))


%% define blur
D_modeled = 9;     % diameter of the modeled blurring kernel
sigma_w = 0.3;     % low noise level     

 
%% ideal image 
x = double(imread('../../data/lena256.bmp'));
[Ny,Nx] = size(x);  % ideal image size
X_f = fft2(x);


%% preparation for L2-r-L0 restoration
% using L2-relaxed-L0 deblurring Matlab Toolbox v2.1
% please check the toolbox for details

Niter = 10;  % number of iterations for L2-r-L0 restoration

% Load parameters for L2-r-L0 deblurring with Niter=10:
% All three representations are active.
% From Table I, ConDy 10 (number of iterations) in Portilla TIP2015,
% parameters jointly optimizing SSIM of the 24 test experiments.
load L2rL0_Niter10_parameters.mat

LB = 0;
UB = 255;


%% loop for different percentages of mismatch
mis_percent_list = [];
SNR_SPA_list = [];
SNR_SPA_ORACLE_list = [];
SNR_IR_list = [];
SNR_IR_ORACLE_list = [];
SNR_simu_list = [];
SNR_simu_ORACLE_list = [];

[modeled_psf_f_o,modeled_psf_f,modeled_kernel,Ly,Lx] = disk_kernel(D_modeled,Ny,Nx,0,0);  % modeled PSF

for mis_percent = 0.6:0.02:1.4   % +-40% size mismatch

    mis_percent_list = [mis_percent_list; mis_percent];

    %% simulate the observation
    D_simu = D_modeled * mis_percent;  % diameter of the blurring kernel for simulation 
    
    [real_psf_f_o,real_psf_f,kernel,Ly,Lx] = disk_kernel(D_simu,Ny,Nx,0,0);  % real PSF

    randn('seed',0);
    Y_f = real_psf_f_o.*X_f + fft2(sigma_w*randn(Ny,Nx)); % observation model for noisy convolved image
    y_simu = real(ifft2(Y_f));
    %figure(1),imagescTSc(y_simu);title('simulation in HR')
    
    
    %% simulated observation by sub-sampling
    y_obsv = y_simu(1:2:end,1:2:end);
    %figure(2),imagescTSc(y_obsv);title('simulated observation in LR')
    
    
    %% get PSD of the uncorrputed image
    % sigma_x and rho (adapted to the [0 255] range) work well for Wiener 
    % (both boundary processing and Wiener "optimized" at the same time). 
    im_power = ceil(log(max(y_obsv(:)))/log(2));
    sigma_x = 30*2^im_power/256;    
    sig2x = sigma_x^2;
    rho = 0.65;

    [xc, yc] = meshgrid(1:Nx,1:Ny); 
    xc = xc - 1 - Nx/2;
    yc = yc - 1 - Ny/2;
    u = xc/Nx;
    v = yc/Ny;
    PX_f = sig2x*4*log(rho)^2./((log(rho)^2 + 4*pi^2*u.^2).*(log(rho)^2 + 4*pi^2*v.^2));
    PX_f = fftshift(PX_f);
    
    
    %%  ******************************************************************************
    % ********     SPA with modeled PSF + restoration for aliasing handling    *******
    % ********************************************************************************
    fprintf(1,'#SPA + Restoration (with modeled PSF) ...\n');
    
    % mask processing
    mask = zeros(Ny,Nx);
    mask(1:2:end,1:2:end) = 1;
    mask = logical(mask);

    % masking and extanding
    y = zeros(Ny,Nx);
    y(1:2:end,1:2:end) = y_obsv;%.*mask(1:2:end,1:2:end);

    % get initial guess with extrapol.m
    y00_obsv = y;
    mu = mean2(y00_obsv.*mask)/mean2(mask);
    y00 = (y00_obsv - mu).*mask;
    % Initiallize surrounded by zeros
    z00 = mu*ones(Ny,Nx);
    z00(mask) = y00_obsv(mask);
    z0 = extrapol(z00, mask);
    
    % SPA
    z_SPA = SPA(z0, modeled_kernel, mask,PX_f, sigma_w);
    %figure(3),imagescTSc(z_SPA);title('result of SPA interpolation')
    
    % L2-rL0 resotration
    x_hat_SPA = deblur_L2relaxedL0(z_SPA, rep, modeled_psf_f_o, sigma_w, Niter, alpha_v, sig2r0_v, beta, LB, UB);
    %figure(4),imagescTSc(x_hat_SPA);title('result of deconvolution on SPA')  
    
    
    %%  *******************************************************************
    % ********     SPA with real PSF + restoration (SPA oracle)     *******
    % *********************************************************************
   
    fprintf(1,'#SPA Oracle + Restoration (with real PSF) ...\n');
    % SPA ORACLE
    z_SPA_ORACLE = SPA(z0, kernel, mask, PX_f, sigma_w);
    %figure(5),imagescTSc(z_SPA_ORACLE);title('result of SPA oracle interploation')
    
    % L2-rL0 resotration
    x_hat_SPA_ORACLE = deblur_L2relaxedL0(z_SPA_ORACLE, rep, real_psf_f_o, sigma_w, Niter, alpha_v, sig2r0_v, beta, LB, UB);
    %figure(6),imagescTSc(x_hat_SPA_ORACLE);title('result of deconvolution on SPA oracle')  
    
    
    %% *****************************************************************************************
    % **************       Interpolation + Restoration (I+R) with modeled PSF       ************
    % **********************************************************************%*******************
    fprintf(1,'# Spline Interpolation + Restoration (with modeled PSF)   ...\n');

    % preparation for spline interpolation
    [Nsy,Nsx] = size(y_obsv);
    [xsc, ysc] = meshgrid(1:Nsx,1:Nsy);
    [xc, yc] = meshgrid(1:0.5:Nsx+1,1:0.5:Nsy+1);
    xc = xc(1:end-1,1:end-1);
    yc = yc(1:end-1,1:end-1);
    
    % #############    INTERPOLATION   #############
    z_IR = interp2(xsc, ysc,double(y_obsv), xc, yc, 'spline');
    %figure(7),imagescTSc(z_IR);title('result of spline interploation')

    % #############     RESTORATION    ##############
    x_hat_IR = deblur_L2relaxedL0(z_IR, rep, modeled_psf_f_o, sigma_w, Niter, alpha_v, sig2r0_v, beta, LB, UB);
    %figure(8),imagescTSc(x_hat_IR);title('result of I + R')
    
    
    %% *****************************************************************************************
    % *******       Interpolation + Restoration (I+R) with real PSF   (IR Oracle)       ********
    % **********************************************************************%*******************
    fprintf(1,'# Spline Interpolation + Restoration (with real PSF)   ...\n');
    
    x_hat_IR_ORACLE = deblur_L2relaxedL0(z_IR, rep, real_psf_f_o, sigma_w, Niter, alpha_v, sig2r0_v, beta, LB, UB);
    %figure(9),imagescTSc(x_hat_IR_ORACLE);title('result of I + R oracle')
    
    
    %%  **************************************************************************************************
    % ********           directly restore full simulation with modeled (reference) PSF            ********
    % ****************************************************************************************************
    fprintf(1,'# Resotration on simultation (with modeled PSF)   ...\n');
    x_hat_simu = deblur_L2relaxedL0(y_simu, rep, modeled_psf_f_o, sigma_w, Niter, alpha_v, sig2r0_v, beta, LB, UB);
    %figure(10),imagescTSc(x_hat_simu);title('result of direct deconvolution on simulation with modeled PSF') 
    
    
    %%  *****************************************************************************************
    % ********           directly restore full simulation with real PSF (oracle)         ********
    % *******************************************************************************************
    fprintf(1,'# Proecessing for oracle (with modeled PSF)   ...\n');
    x_hat_simu_ORACLE = deblur_L2relaxedL0(y_simu, rep, real_psf_f_o, sigma_w, Niter, alpha_v, sig2r0_v, beta, LB, UB);
    %figure(11),imagescTSc(x_hat_simu_ORACLE);title('result of direct deconvolution on simulation with real PSF (oracle)') 
    
    
    %% calculate SNR
    SNR_SPA = 10*log10(mean2(x.^2)/mean2((x_hat_SPA - x).^2)) 
    SNR_IR = 10*log10(mean2(x.^2)/mean2((x_hat_IR - x).^2)) 
    SNR_simu = 10*log10(mean2(x.^2)/mean2((x_hat_simu - x).^2)) 
    
    SNR_SPA_list = [SNR_SPA_list; SNR_SPA];
    SNR_IR_list = [SNR_IR_list; SNR_IR];
    SNR_simu_list = [SNR_simu_list; SNR_simu];
    

end


%% plot
figure(12)
plot(mis_percent_list,SNR_simu_list,'ro-',mis_percent_list,SNR_SPA_list,'bo-',mis_percent_list,SNR_IR_list,'ko-'); 
axis([0.6 1.4 0 26]); grid on; hold on
legend('not aliased (wo/ SPA)', 'aliased (w/ SPA)', 'aliased (wo/ SPA)');
title('Performance effect of the PSF size mismatch');
set(gca,'fontsize',20);

h = gca;
set(gcf,'PaperUnits','inches','PaperPosition',[0 0 7 8])
set(h, 'XTick', [0.6:0.1:1.4]);

h = xlabel('Mismatch factor (1=no mismatch)');
set(h,'FontSize',20);
h = ylabel('SNR(dB)');
set(h,'FontSize',20);
