% REPORDUCE_TABLE_1:
%
% Code to reproduce all experiments in IPOL submission section 3.1 
% Non-circulant Deconvolution and ISNR data in Table 1.
%
% Requires the L2-r-L0 Deblur Matlab Toolbox v2.1:
% https://www.researchgate.net/publication/325903515_L2rL0deblurMatlabToolbox2p1
% It is included in this ZIP file.
% It is an implementation of the method described in:
% - J. Portilla, A. Tristan-Vega, I.W. Selesnick, "Efficient and robust
%   image restoration using multiple-feature L2-relaxed sparse analysis
%   priors", IEEE Transactions on Image Processing, vol. 24, no. 12,
%   pp. 5046-5059, Dec 2015.
%
% % Includes the original MATLAB function "edgetaper" renamed as "edgetaper_matlab" 
% for execution by Octave. The authors claim no rights to this function. 
%
% IPOL SUBMISSION "Spectral Pre-Adaptation for Restoring Real-World Blurred 
% Images Using Standard Deconvolution Methods"
%
% Version 1.2, June 2022
%
% Copyright (c) 2022
% Chaoqun Dong <chaodong999@gmail.com>
% Javier Portilla <javier.portilla@csic.es>
%
% This program is free software: you can redistribute it and/or modify
% it under the terms of the GNU Affero General Public License as
% published by the Free Software Foundation, either version 3 of the
% License, or (at your option) any later version.
%
% This program is distributed in the hope that it will be useful,
% but WITHOUT ANY WARRANTY; without even the implied warranty of
% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
% GNU Affero General Public License for more details.
%
% You should have received a copy of the GNU Affero General Public License
% along with this program. If not, see <http://www.gnu.org/licenses/>.


%%
clear all;
close all;

addpath(genpath('../../data'))
addpath(genpath('../../core'))
addpath(genpath('../../functions'))
addpath(genpath('../../L2rL0deblurMatlabToolbox2p1'))


%% set up
% We tested SPA for dealing with non-circulant boundary condition with
% combinations of 3 different images, 4 different blurring kernels and
% 2 different noise levels (8 different degrations in total).
% We also tested with two different restoration methods, Wiener filter
% and L2-r-L0.
%
%     img       - input image
%                 img = 'house'         - use House image,
%                 img = 'cameraman'     - use Cameraman image
%                 img = 'straw'         - use Straw image
%     deg       - the degradation
%                 deg = 1     - PSF1 + low noise (sigma_w.^2 = 0.25)
%                 deg = 2     - PSF1 + medium noise (sigma_w.^2 = 2)
%                 deg = 3     - PSF2 + low noise (sigma_w.^2 = 0.31)
%                 deg = 4     - PSF2 + medium noise (sigma_w.^2 = 4)
%                 deg = 5     - PSF3 + low noise (sigma_w.^2 = 1)
%                 deg = 6     - PSF3 + medium noise (sigma_w.^2 = 4)
%                 deg = 7     - PSF4 + low noise (sigma_w.^2 = 0.25)
%                 deg = 8     - PSF4 + medium noise (sigma_w.^2 = 4)
%                 (PSF1, out of focus; PSF2, uniform; PSF3, motion blur; PSF4, oblique)
%     restor    - restoration method
%                 restor = 'Wiener'    - using Wiener filtering
%                 restor = 'L2rL0'     - using L2-relaxed-L0 restoration


isnr_matrix = [];  % data in IPOL submission Table 1

% define kernels
% PSF1, non-separable kernel
i = meshgrid(-7:7, -7:7); j = i';
kernel = 1./(i.^2 + j.^2 + 1);
kernel = kernel/sum(kernel(:));
PSF1 = kernel;
% PSF2, uniform blur
kernel = ones(9)/81;
kernel = kernel/sum(kernel(:));
PSF2 = kernel;
% PSF3, motion blur
kernel = ones(9,1);
kernel = kernel/sum(kernel(:));
PSF3 = kernel;
% PSF4, oblique blur, non-separable kernel
kernel = [0 0 0 0 0 0 0;
    0 0 0 0 1 1 1;
    0 0 1 2 3 2 1;
    0 1 3 4 3 1 0;
    1 2 3 2 1 0 0;
    1 1 1 0 0 0 0;
    0 0 0 0 0 0 0];
kernel = kernel/sum(kernel(:));
PSF4 = kernel;

for res = 1:2
    if res == 1,
        restor = 'Wiener';
    elseif res ==2,
        restor = 'L2rL0';
    end
    
    tt = 0;
    
    for nim = 1:3
        for deg = 1:8
            
            tt = tt+1;   % ISNR table
            
            % read image
            if nim == 1,
                x = imread('../../data/house.png');
                im_name = 'House';
            elseif nim==2,
                x = imread('../../data/cameraman.png');
                im_name = 'Cameraman';
            elseif nim==3,
                x = imread('../../data/straw.tif');
                im_name = 'Straw';
            end
            
            
            % define degradations
            if deg==1,
                kernel = PSF1;
                sigma_w = sqrt(0.25);
            elseif deg==2,
                kernel = PSF1;
                sigma_w = sqrt(2);
            elseif deg==3,
                kernel = PSF2;
                sigma_w = 0.56; %sqrt(0.308);
            elseif deg==4,
                kernel = PSF2;
                sigma_w = sqrt(4);
            elseif deg==5,
                kernel = PSF3;
                sigma_w = sqrt(1);
            elseif deg==6,
                kernel = PSF3;
                sigma_w = sqrt(4);
            elseif deg==7,
                kernel = PSF4;
                sigma_w = sqrt(0.25);
            elseif deg==8,
                kernel = PSF4;
                sigma_w = sqrt(4);
            end
            
            fprintf(1,'*** Restoration: %s, Test image: %s, Degradation #%d ***\n', restor, im_name, deg);
            
            %% simulate observation
            [Ny,Nx] = size(x);   % ideal image size
            
            % make sure the kernel is normalized
            kernel = kernel/sum(sum(kernel));
            [ny,nx] = size(kernel);  % in our implementation the kernel size should be odd for both dimensions
            % half kernel size
            Ly = ceil((ny-1)/2);
            Lx = ceil((nx-1)/2);
            
            % Fourier transform
            h = zeros(Ny,Nx);
            h(Ny/2+1-Ly:Ny/2+1+Ly,Nx/2+1-Lx:Nx/2+1+Lx) = kernel;
            H_f= fft2(fftshift(h));
            
            x = double(x);
            X_f = fft2(x);
            randn('seed',0);
            Y_f = H_f.*X_f + fft2(sigma_w*randn(size(x)));
            y_simu = real(ifft2(Y_f));
            
            y_obsv = y_simu(Ly+1:end-Ly,Lx+1:end-Lx);  % cut off the invalid boundary pixels
            
            %figure(1),imagescTSc(y_obsv);title('simulated observation')
            
            
            %% set extra boundary extension size
            [Ley, Lex] = get_extra_size(kernel, Ny, Nx);
            
            % final extended size
            NDy = Ny + 2*Ley;
            NDx = Nx + 2*Lex;
            
            %% psf in final extended size
            h = zeros(NDy,NDx);
            h(NDy/2+1-Ly:NDy/2+1+Ly,NDx/2+1-Lx:NDx/2+1+Lx) = kernel;
            D_f= fft2(fftshift(h));
            
            
            %% get PSD of the uncorrputed image
            % sigma_x and rho (adapted to the [0 255] range) work well for Wiener
            % (both boundary processing and Wiener "optimized" at the same time).
            im_power = double(ceil(log(max(y_obsv(:)))/log(2)));
            im_range = 2^im_power;
            PX_f = get_x_psd(im_range, NDy, NDx);
            
            
            %% preparation for Wiener filter
            % blurred noisy image PSD
            PY_f = abs(D_f).^2.*PX_f + sigma_w^2;
            % Wiener filter
            Hwie_f = PX_f.*conj(D_f)./PY_f;
            
            
            %% preparation for L2-relaxed-L0 restoration
            % using L2-relaxed-L0 deblurring Matlab Toolbox v2.1
            % please check the toolbox for details
            
            Niter = 10;  % number of iterations for L2-r-L0 restoration
            
            % Load parameters for L2-r-L0 deblurring with Niter=10:
            % All three representations are active.
            % From Table I, CoNy 10 (number of iterations) in Portilla TIP2015,
            % parameters jointly optimizing SSIM of the 24 test experiments.
            load L2rL0_Niter10_parameters.mat
            
            LB = 0;
            UB = 255;
            
            
            %% *****************************************************************************************
            % ********     Method One: boundary mirror extension and edgetaper tiling (ET)      ********
            % ********                     +     Restoration                                    ********
            % ******************************************************************************************
            fprintf(1,'# Method One: ET + Restoration  ...\n');
            
            
            % #############       ET      ###############
            
            z_ET = bound_extension(y_obsv,Ly+Ley,Lx+Lex,'mirror'); % mirror extension
            PSF = fspecial('gaussian', 40, 6);
            % z_ET  = edgetaper(z_ET, PSF);   % for MATLAB
            z_ET  = edgetaper_matlab(z_ET, PSF);  % for Octave
            
            %figure(2),imagescTSc(z_ET);title('result of ET')
            
            
            % #############     RESTORATION    #############
            if strcmp(restor,'Wiener')
                
                % ET
                Z_f_ET = fft2(z_ET);
                x_hat_ET = real(ifft2(Z_f_ET.*Hwie_f));
                x_hat_ET = max(min(x_hat_ET, im_range-1), 0);  % back to image range
                
            elseif strcmp(restor,'L2rL0')
                
                fprintf(1,'Info:  Deblurring ET result  ...\n');
                x_hat_ET = deblur_L2relaxedL0(z_ET, rep, D_f, sigma_w, Niter, alpha_v, sig2r0_v, beta, LB, UB);
                
            end
            
            % cut off the extras extended pixls
            % restoration has the same size as ideal image
            x_hat_ET = x_hat_ET(Ley+1:end-Ley,Lex+1:end-Lex);
            
            %figure(3),imagescTSc(x_hat_ET);title('result of deconvolution on ET')
            
            
            %% ******************************************************************************************
            % *********************           Method Two: WML + Restoration           *******************
            % *******************************************************************************************
            fprintf(1,'# Method Two: WML + Restoration  ...\n');

            %  #############    wrap_boundry_liu.m    #############
            y_obsv_size = size(y_obsv);
            WML_img_size = y_obsv_size*2;  % twice the size of the observation
            z_WML = wrap_boundary_liu(y_obsv/255, WML_img_size);
           
            %figure(4),imagescTSc(z_WML);title('result of WML')


            % #############     RESTORATION    #############
            % preparation for Wiener filter with new size
            % psf in final extended size
            NDy_WML = WML_img_size(1);
            NDx_WML = WML_img_size(2);
            h = zeros(NDy_WML,NDx_WML);
            h(NDy_WML/2+1-Ly:NDy_WML/2+1+Ly,NDx_WML/2+1-Lx:NDx_WML/2+1+Lx) = kernel;
            D_f_WML= fft2(fftshift(h));

            % get PSD of the uncorrputed image
            PX_f_WML = get_x_psd(im_range, NDy_WML, NDx_WML);
       
            % blurred noisy image PSD
            PY_f_WML = abs(D_f_WML).^2.*PX_f_WML + sigma_w^2;
            % Wiener filter
            Hwie_f_WML = PX_f_WML.*conj(D_f_WML)./PY_f_WML;

            z_WML = z_WML*255;
            if strcmp(restor,'Wiener')

                Z_f_WML = fft2(z_WML);
                x_hat_WML = real(ifft2(Z_f_WML.*Hwie_f_WML));
                x_hat_WML = max(min(x_hat_WML, im_range-1), 0);  % back to image range

            elseif strcmp(restor,'L2rL0')

                fprintf(1,'Info:  Deblurring WML result  ...\n');
                x_hat_WML = deblur_L2relaxedL0(z_WML, rep, D_f_WML, sigma_w, Niter, alpha_v, sig2r0_v, beta, LB, UB);

            end

            % cut off the extras extended pixls
            % restoration has the same size as ideal image
            Ey_WML = NDy_WML - y_obsv_size(1);
            Ex_WML = NDx_WML - y_obsv_size(2);
            x_hat_WML = x_hat_WML(1:end-Ey_WML,1:end-Ex_WML);
            
            %figure(5),imagescTSc(x_hat_WML);title('result of deconvolution on WML')
            
            
            %% ******************************************************************************************
            % *********************           Method Three: SPA + Restoration           *******************
            % *******************************************************************************************
            fprintf(1,'# Method Two: SPA + Restoration  ...\n');
            
            
            % #############      SPA     #############
            
            % mask for SPA
            mask = zeros(NDy,NDx);
            mask(Ley+Ly+1:end-Ly-Ley,Lex+Lx+1:end-Lx-Lex) = 1;
            mask = logical(mask);
            
            % masking and extending
            y = zeros(NDy,NDx);
            y(Ley+Ly+1:end-Ly-Ley,Lex+Lx+1:end-Lx-Lex) = y_obsv.*mask(Ley+Ly+1:end-Ly-Ley,Lex+Lx+1:end-Lx-Lex);
            
            % get initial guess with extrapol.m
            y00_obsv = y;
            mu = mean2(y00_obsv.*mask)/mean2(mask);
            y00 = (y00_obsv - mu).*mask;
            % Initiallize surrounded by zeros
            z00 = mu*ones(NDy,NDx);
            z00(mask) = y00_obsv(mask);
            z0 = extrapol(z00, mask);
            
            % SPA
            noiter = 50;
            z_SPA = SPA(z0,kernel,mask,PX_f,sigma_w,noiter);
            
            %figure(6),imagescTSc(z_SPA);title('result of SPA')
            
            
            % #############     RESTORATION    #############
            if strcmp(restor,'Wiener')
                
                Z_f_SPA = fft2(z_SPA);
                x_hat_SPA = real(ifft2(Z_f_SPA.*Hwie_f));
                x_hat_SPA = max(min(x_hat_SPA, im_range-1), 0);  % back to image range
                
            elseif strcmp(restor,'L2rL0')
                
                fprintf(1,'Info:  Deblurring SPA result  ...\n');
                x_hat_SPA = deblur_L2relaxedL0(z_SPA, rep, D_f, sigma_w, Niter, alpha_v, sig2r0_v, beta, LB, UB);
                
            end
            
            % cut off the extras extended pixels
            % restoration has the same size as ideal image
            x_hat_SPA = x_hat_SPA(Ley+1:end-Ley,Lex+1:end-Lex);
            
            %figure(7),imagescTSc(x_hat_SPA);title('result of deconvolution on SPA')
            
            
            %% Oracle
            % directly deblurring on the simulation
            fprintf(1,'Processing for Oracle  ...\n');
            
            
            if strcmp(restor,'Wiener')
                
                % PSD of the uncorrputed image in orginal size
                PX_f = get_x_psd(im_range, Ny, Nx);
                
                PY_f = abs(H_f).^2.*PX_f + sigma_w^2;
                Hwie_f = PX_f.*conj(H_f)./PY_f;
                
                Y_SIMU_f = fft2(y_simu);
                x_hat_Ora = real(ifft2(Y_SIMU_f.*Hwie_f));
                x_hat_Ora = max(min(x_hat_Ora, im_range-1), 0);  % back to image range
                
            elseif strcmp(restor,'L2rL0')
                
                fprintf(1,'Info:  Deblurring on simulation to get Oracle result  ...\n');
                x_hat_Ora = deblur_L2relaxedL0(y_simu, rep, H_f, sigma_w, Niter, alpha_v, sig2r0_v, beta, LB, UB);
                
            end
            
            %figure(8),imagescTSc(x_hat_Ora);title('Oracle')
            
            
            %% ISNR
            % compute ISNR without the invalid boundary pixels affected by CBC
            psnr_simu = psnr(y_simu(Ly+1:end-Ly,Lx+1:end-Lx), x(Ly+1:end-Ly,Lx+1:end-Lx));
            
            psnr_ET = psnr(x_hat_ET(Ly+1:end-Ly,Lx+1:end-Lx), x(Ly+1:end-Ly,Lx+1:end-Lx));
            isnr_ET = psnr_ET - psnr_simu
            
            psnr_WML = psnr(x_hat_WML, x(Ly+1:end-Ly,Lx+1:end-Lx));
            isnr_WML = psnr_WML - psnr_simu
            
            psnr_SPA = psnr(x_hat_SPA(Ly+1:end-Ly,Lx+1:end-Lx), x(Ly+1:end-Ly,Lx+1:end-Lx));
            isnr_SPA = psnr_SPA - psnr_simu
            
            psnr_Ora = psnr(x_hat_Ora(Ly+1:end-Ly,Lx+1:end-Lx),x(Ly+1:end-Ly,Lx+1:end-Lx));
            isnr_Ora = psnr_Ora - psnr_simu
            
            if res == 1,
                isnr_matrix(tt,1) = isnr_ET;
                isnr_matrix(tt,2) = isnr_WML;
                isnr_matrix(tt,3) = isnr_SPA;
                isnr_matrix(tt,4) = isnr_Ora;
            elseif res ==2,
                isnr_matrix(tt,5) = isnr_ET;
                isnr_matrix(tt,6) = isnr_WML;
                isnr_matrix(tt,7) = isnr_SPA;
                isnr_matrix(tt,8) = isnr_Ora;
            end
        end
    end
    
end

% round ISNR to 2 decimal place
isnr_table = isnr_matrix;
[r,c] = size(isnr_table);
for i = 1:r,
    for j = 1:c
        isnr_table(i,j) = round(isnr_table(i,j)*100)/100;
    end
end

% average
avg = zeros(1,c);
avg_r = zeros(1,c);
for i = 1:c
    avg(i) = mean(isnr_matrix(:,i));
    avg_r(i) = round(avg(i)*100)/100;
end
